# encoding: utf-8
# Created by Jeremy Bowman on Fri Feb 21 11:15:41 EST 2014
# Copyright (c) 2014, 2015 Safari Books Online. All rights reserved.
#
# This software may be modified and distributed under the terms
# of the 3-clause BSD license.  See the LICENSE file for details.
"""
Custom Django settings for Yet Another Django Profiler middleware
"""

from __future__ import unicode_literals

from importlib import import_module
import os
import platform
import re
import sys

from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import cached_property
from django.utils.translation import ugettext as _

DEFAULT_MODULE_PARENT_DIR_PATTERNS = [
    r'\.egg[/\\]',
    r'site-packages[/\\]',
    r'python\d+\.\d+[/\\]',
]


def path_to_module_name(path):
    """Get the fully qualified module name for the given absolute path by
    making intelligent guesses about the PYTHONPATH."""
    if path.endswith('.py'):
        path = path[:-3]
    cwd = os.getcwd()
    if path.startswith(cwd):
        path = path[len(cwd) + 1:]
    for pattern in settings.YADP_MODULE_PARENT_DIR_PATTERNS:
        match = True
        while match:
            match = re.search(pattern, path)
            if match:
                path = path[match.end():]
    path = path.replace('/', '.')
    path = path.replace('\\', '.')
    return path


class LazySettings(object):

    @property
    def YADP_ENABLED(self):
        return getattr(django_settings, 'YADP_ENABLED', django_settings.DEBUG)

    @property
    def YADP_FRACTION_PARAMETER(self):
        return getattr(django_settings, 'YADP_FRACTION_PARAMETER', 'fraction')

    @property
    def YADP_MAX_CALLS_PARAMETER(self):
        return getattr(django_settings, 'YADP_MAX_CALLS_PARAMETER', 'max_calls')

    @property
    def YADP_MODULE_PARENT_DIR_PATTERNS(self):
        return getattr(django_settings, 'YADP_MODULE_PARENT_DIR_PATTERNS',
                       DEFAULT_MODULE_PARENT_DIR_PATTERNS)

    @property
    def YADP_PATH_TO_MODULE_FUNCTION(self):
        return getattr(django_settings, 'YADP_PATH_TO_MODULE_FUNCTION',
                       'yet_another_django_profiler.conf.path_to_module_name')

    @property
    def YADP_PATTERN_PARAMETER(self):
        return getattr(django_settings, 'YADP_PATTERN_PARAMETER', 'pattern')

    @property
    def YADP_PROFILE_PARAMETER(self):
        return getattr(django_settings, 'YADP_PROFILE_PARAMETER', 'profile')

    @property
    def YADP_CLOCK_PARAMETER(self):
        return getattr(django_settings, 'YADP_CLOCK_PARAMETER', 'clock')

    @property
    def YADP_PROFILER_BACKEND(self):
        backend = getattr(django_settings, 'YADP_PROFILER_BACKEND', 'cProfile')
        if backend == 'yappi':
            msg = _('Yappi does not work on {platform}, please select a different value for YADP_PROFILER_BACKEND')
            if platform.python_implementation() == 'PyPy':
                raise ImproperlyConfigured(msg.format('PyPy'))
            version = sys.version_info[:2]
            if version == (3, 2):
                raise ImproperlyConfigured(msg.format('Python 3.2'))
            if version == (3, 1):
                raise ImproperlyConfigured(msg.format('Python 3.1'))
        return backend

    @cached_property
    def path_to_module_function(self):
        """Resolve the function to use for converting paths to module names"""
        mod_name, function_name = self.YADP_PATH_TO_MODULE_FUNCTION.rsplit('.', 1)
        mod = import_module(mod_name)
        return getattr(mod, function_name)


settings = LazySettings()



# encoding: utf-8
# Created by Jeremy Bowman on Mon Feb  1 12:48:09 EST 2016
# Copyright (c) 2016 Safari Books Online. All rights reserved.
#
# This software may be modified and distributed under the terms
# of the 3-clause BSD license.  See the LICENSE file for details.
"""
Utility functions for Yet Another Django Profiler
"""

import marshal
import os
import subprocess
import sys
import tempfile


def run_gprof2dot(profiler):
    """Run gprof2dot.py on the data from the provided profiler to generate a
    call graph in PDF format.

    :param profiler The profiler containing the data to be graphed
    :return A tuple of the subprocess return code and its stdout (the PDF data)
    """
    with tempfile.NamedTemporaryFile() as stats:
        stats.write(marshal.dumps(profiler.stats))
        stats.flush()
        cmd = ('gprof2dot.py -f pstats {} | dot -Tpdf'.format(stats.name))
        env = os.environ.copy()
        env['PYTHONPATH'] = os.pathsep.join(sys.path)
        process = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE, env=env)
        output = process.communicate()[0]
        return_code = process.poll()
    return return_code, output



# encoding: utf-8
# Created by Jeremy Bowman on Fri Feb 21 11:16:36 EST 2014
# Copyright (c) 2014, 2015 Safari Books Online. All rights reserved.
#
# This software may be modified and distributed under the terms
# of the 3-clause BSD license.  See the LICENSE file for details.
"""
Yet Another Django Profiler "profile" management command
"""

from __future__ import unicode_literals

import argparse
import atexit
import cProfile
from optparse import make_option
import pstats
import sys

try:
    from unittest import mock
except ImportError:
    import mock

import django
from django.conf import settings as django_settings
from django.core.management import call_command, ManagementUtility
from django.core.management.base import BaseCommand
from django.utils.six.moves import cStringIO as StringIO

from yet_another_django_profiler.conf import settings
from yet_another_django_profiler.middleware import func_strip_path, which
from yet_another_django_profiler.utils import run_gprof2dot

OPTIONS = (
    (('-o', '--output'), {'dest': 'path', 'help': 'Path to a file in which to store the profiling output (required if generating a call graph PDF, other results are output to the console by default)'}),
    (('-s', '--sort'), {'dest': 'sort', 'help': 'Statistic by which to sort the profiling data (default is to generate a call graph PDF instead)'}),
    (('-f', '--fraction'), {'dest': 'fraction', 'help': 'The fraction of total function calls to display (the default of .2 is omitted if max-calls or pattern are specified)'}),
    (('-m', '--max-calls'), {'dest': 'max_calls', 'help': 'The maximum number of function calls to display'}),
    (('-p', '--pattern'), {'dest': 'pattern', 'help': 'Regular expression filter for function display names'}),
    (('-b', '--backend'), {'dest': 'backend', 'default': 'cProfile', 'help': 'Profiler backend to use (cProfile or yappi)'}),
    (('-c', '--clock'), {'dest': 'clock', 'default': 'cpu', 'help': 'Yappi clock type to use (cpu or wall)'}),
)


class Command(BaseCommand):
    """
    Django management command for profiling other management commands.
    """
    args = 'other_command <argument argument ...>'
    help = 'Profile another Django management command'
    # Command line arguments for Django 1.7 and below
    custom_options = tuple([make_option(*option[0], **option[1]) for option in OPTIONS])

    @property
    def use_argparse(self):
        return not (django.VERSION[0] == 1 and django.VERSION[1] < 8)

    def add_arguments(self, parser):
        """Command line arguments for Django 1.8+"""
        for option in OPTIONS:
            parser.add_argument(*option[0], **option[1])
        parser.add_argument('other_command', help='The management command to be profiled')
        parser.add_argument(
            'command_arguments', nargs=argparse.REMAINDER, help='Arguments of the management command being profiled')

    def create_parser(self, prog_name, subcommand):
        """
        Override the base create_parser() method to ignore options of the
        command being profiled.
        """
        if not self.use_argparse:
            self.__class__.option_list = BaseCommand.option_list + self.custom_options
        parser = super(Command, self).create_parser(prog_name, subcommand)
        if not self.use_argparse:
            parser.disable_interspersed_args()
        return parser

    def handle(self, *args, **options):
        """
        Run and profile the specified management command with the provided
        arguments.
        """
        if not self.use_argparse and not len(args):
            self.print_help(sys.argv[0], 'profile')
            sys.exit(1)
        if not options['sort'] and not options['path']:
            self.stdout.write('Output file path is required for call graph generation')
            sys.exit(1)

        if self.use_argparse:
            command_name = options['other_command']
        else:
            command_name = args[0]
        utility = ManagementUtility(sys.argv)
        command = utility.fetch_command(command_name)
        parser = command.create_parser(sys.argv[0], command_name)
        if self.use_argparse:
            command_options = parser.parse_args(options['command_arguments'])
            command_args = vars(command_options).pop('args', ())
        else:
            command_options, command_args = parser.parse_args(list(args[1:]))

        if command_name == 'test' and django_settings.TEST_RUNNER == 'django_nose.NoseTestSuiteRunner':
            # Ugly hack: make it so django-nose won't have nosetests choke on
            # our parameters
            BaseCommand.option_list += self.custom_options

        if options['backend'] == 'yappi' or (settings.YADP_PROFILER_BACKEND == 'yappi' and not options['backend']):
            import yet_another_django_profiler.yadp_yappi as yadp_yappi
            profiler = yadp_yappi.YappiProfile(wall=options['clock'] == 'wall')
        else:
            profiler = cProfile.Profile()

        if 'testing' not in options:
            atexit.register(output_results, profiler, options, self.stdout)
        profiler.runcall(call_command, command_name, *command_args, stderr=self.stderr,
                         stdout=self.stdout, **vars(command_options))
        if 'testing' in options:
            output_results(profiler, options, self.stdout)
        else:
            sys.exit(0)


def output_results(profiler, options, stdout):
    """Generate the profiler output in the desired format.  Implemented as a
    separate function so it can be run as an exit handler (because management
    commands often call exit() directly, bypassing the rest of the profile
    command's handle() method)."""
    profiler.create_stats()

    if not options['sort']:
        if not which('dot'):
            stdout.write('Could not find "dot" from Graphviz; please install Graphviz to enable call graph generation')
            return
        if not which('gprof2dot.py'):
            stdout.write('Could not find gprof2dot.py, which should have been installed by yet-another-django-profiler')
            return
        return_code, output = run_gprof2dot(profiler)
        if return_code:
            stdout.write('gprof2dot/dot exited with {}'.format(return_code))
            return
        path = options['path']
        with open(path, 'wb') as pdf_file:
            pdf_file.write(output)
            stdout.write('Wrote call graph to {}'.format(path))
    else:
        sort = options['sort']
        if sort == 'file':
            # Work around bug on Python versions >= 2.7.4
            sort = 'fil'
        out = StringIO()
        stats = pstats.Stats(profiler, stream=out)
        with mock.patch('pstats.func_strip_path') as mock_func_strip_path:
            mock_func_strip_path.side_effect = func_strip_path
            stats.strip_dirs()
        restrictions = []
        if options['pattern']:
            restrictions.append(options['pattern'])
        if options['fraction']:
            restrictions.append(float(options['fraction']))
        elif options['max_calls']:
            restrictions.append(int(options['max_calls']))
        elif not options['pattern']:
            restrictions.append(.2)
        stats.sort_stats(sort).print_stats(*restrictions)
        if options['path']:
            path = options['path']
            with open(path, 'w') as text_file:
                text_file.write(out.getvalue())
                stdout.write('Wrote profiling statistics to {}'.format(path))
        else:
            stdout.write(out.getvalue())



# encoding: utf-8
# Created by Jeremy Bowman on Fri Feb 21 11:16:36 EST 2014
# Copyright (c) 2014, 2015 Safari Books Online. All rights reserved.
#
# This software may be modified and distributed under the terms
# of the 3-clause BSD license.  See the LICENSE file for details.
"""
Yet Another Django Profiler middleware implementation
"""

from __future__ import unicode_literals

import cProfile
import logging
import os
import pstats

try:
    from unittest import mock
except ImportError:
    import mock

from django.core.exceptions import MiddlewareNotUsed
from django.utils.six.moves import cStringIO as StringIO
from django.utils.translation import ugettext as _

from .conf import settings
from .utils import run_gprof2dot

log = logging.getLogger(__name__)


def func_strip_path(func_name):
    """Replacement for pstats.func_strip_path which yields qualified module names"""
    filename, line, name = func_name
    return settings.path_to_module_function(filename), line, name


def in_request(request, parameter):
    """Determine if the request contains the specified parameter, whether it's
    a GET or a POST."""
    return parameter in request.GET or parameter in request.POST


def set_content(response, content):
    """Set the content of the provided response, whether or not it's streaming"""
    if response.streaming:
        # No point in consuming the previous iterator, the profiler has
        # already stopped
        response.streaming_content = [content]
    else:
        response.content = content


def text_response(response, content):
    """Return a plain text message as the response content."""
    set_content(response, content)
    response['Content-type'] = 'text/plain'
    return response


def which(program):
    """Return the path of the named program in the PATH, or None if no such
    executable program can be found  Used to make sure that required binaries
    are in place before attempting to call them."""

    def is_exe(file_path):
        return os.path.isfile(file_path) and os.access(file_path, os.X_OK)

    program_path, _name = os.path.split(program)
    if program_path:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


class ProfilerMiddleware(object):
    """
    Code profiling middleware which can display either a call graph PDF or
    a table of called functions ordered by the desired statistic.  For the call
    graph, just append "?profile" to the URL.  For the graph generation to
    work, install Graphviz from http://www.graphviz.org/Download.php

    For a statistics table, use the statistic you want to sort by as the
    parameter (such as "?profile=time").  Sorting options include:

    * calls (call count)
    * cumulative (cumulative time)
    * file (file name)
    * module (file name)
    * pcalls (primitive call count)
    * line (line number)
    * name (function name)
    * nfl (name/file/line)
    * stdname (standard name)
    * time (internal time)

    Additional parameters can be added when generating a statistics table:

    * fraction - The fraction of total function calls to display (the default of .2 is omitted if max_calls or pattern are specified)
    * max_calls - The maximum number of function calls to display
    * pattern - Regular expression filter for function display names

    To get these instructions in the app if you forget the usage options, use
    "?profile=help" in the URL.

    Inspiration:

    * https://gist.github.com/kesor/1229681
    * https://bitbucket.org/brodie/geordi
    """

    def __init__(self, get_response=None):
        if not settings.YADP_ENABLED:
            # Disable the middleware completely when YADP_ENABLED = False
            raise MiddlewareNotUsed()
        self.get_response = get_response
        self.error = None
        self.profiler = None
        self.get_parameters = None
        self.post_parameters = None
        self.clock_parameter = None
        self.fraction_parameter = None
        self.max_calls_parameter = None
        self.pattern_parameter = None
        self.profile_parameter = None

    def __call__(self, request):
        response = self.get_response(request)
        return self.process_response(request, response)

    def get_parameter(self, name):
        """Get the specified parameter from the request, whether it's a GET or a
        POST, and then removes it so it doesn't interfere with the view's
        normal operation.  Returns `None` if it is not present.
        Search in GET (and stop if found), then in POST."""
        if self.get_parameters is not None and name in self.get_parameters:
            value = self.get_parameters.get(name, None)
            self.get_parameters.pop(name)
            return value

        if self.post_parameters is not None and name in self.post_parameters:
            value = self.post_parameters.get(name, None)
            self.post_parameters.pop(name)
            return value

        return None

    def process_view(self, request, callback, callback_args, callback_kwargs):
        self.profile_parameter = None
        if settings.YADP_ENABLED and in_request(request, settings.YADP_PROFILE_PARAMETER):
            self.error = None
            self.get_parameters = request.GET.copy()
            request.GET = self.get_parameters
            self.post_parameters = request.POST.copy()
            request.POST = self.post_parameters
            self.fraction_parameter = self.get_parameter(settings.YADP_FRACTION_PARAMETER)
            self.max_calls_parameter = self.get_parameter(settings.YADP_MAX_CALLS_PARAMETER)
            self.pattern_parameter = self.get_parameter(settings.YADP_PATTERN_PARAMETER)
            self.profile_parameter = self.get_parameter(settings.YADP_PROFILE_PARAMETER)
            if settings.YADP_PROFILER_BACKEND == 'yappi':
                try:
                    from .yadp_yappi import YappiProfile
                    wall = self.get_parameter(settings.YADP_CLOCK_PARAMETER) == 'wall'
                    self.profiler = YappiProfile(wall=wall)
                except Exception as e:
                    log.exception(e)
                    self.error = _('Could not find Yappi; please install Yappi to be able to use it for profiling')
                    return None
            else:
                self.profiler = cProfile.Profile()
            args = (request,) + callback_args
            return self.profiler.runcall(callback, *args, **callback_kwargs)

    def process_response(self, request, response):
        if self.profile_parameter is not None:
            if self.error:
                return text_response(response, self.error)
            self.profiler.create_stats()
            mode = self.profile_parameter
            if mode == 'file':
                # Work around bug on Python versions >= 2.7.4
                mode = 'fil'
            if not mode:
                if not which('dot'):
                    return text_response(response, _('Could not find "dot" from Graphviz; please install Graphviz to enable call graph generation'))
                if not which('gprof2dot.py'):
                    return text_response(response, _('Could not find gprof2dot.py, which should have been installed by yet-another-django-profiler'))
                return_code, output = run_gprof2dot(self.profiler)
                if return_code:
                    raise Exception(_('gprof2dot.py exited with {return_code}').format(return_code=return_code))
                set_content(response, output)
                response['Content-Type'] = 'application/pdf'
                return response
            elif mode == 'help':
                return text_response(response, ProfilerMiddleware.__doc__)
            else:
                out = StringIO()
                stats = pstats.Stats(self.profiler, stream=out)

                with mock.patch('pstats.func_strip_path') as mock_func_strip_path:
                    mock_func_strip_path.side_effect = func_strip_path
                    stats.strip_dirs()
                restrictions = []
                if self.pattern_parameter is not None:
                    restrictions.append(self.pattern_parameter)
                if self.fraction_parameter is not None:
                    restrictions.append(float(self.fraction_parameter))
                elif self.max_calls_parameter is not None:
                    restrictions.append(int(self.max_calls_parameter))
                elif self.pattern_parameter is None:
                    restrictions.append(.2)
                stats.sort_stats(mode).print_stats(*restrictions)
                return text_response(response, out.getvalue())
        return response



# encoding: utf-8
# Created by Jeremy Bowman on Thu Feb 20 16:54:00 EST 2014
# Copyright (c) 2014, 2015 Safari Books Online. All rights reserved.
#
# This software may be modified and distributed under the terms
# of the 3-clause BSD license.  See the LICENSE file for details.
"""
yet-another-django-profiler setup script
"""

import codecs
import sys
from setuptools import find_packages, setup

from yet_another_django_profiler import __version__

version = '.'.join(str(n) for n in __version__)

install_requires = [
    'Django',
]

if '{0.major}.{0.minor}'.format(sys.version_info) < '3.3':
    install_requires.append('mock')

with codecs.open('README.rst', 'r', 'utf-8') as f:
    long_description = f.read()

setup(
    name='yet-another-django-profiler',
    version=version,
    author='Jeremy Bowman',
    author_email='jbowman@safaribooksonline.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django :: 1.7',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
        'Topic :: Software Development',
    ],
    description='Django middleware for performance profiling directly from the browser',
    long_description=long_description,
    url='http://github.com/safarijv/yet-another-django-profiler',
    packages=find_packages(exclude=['ez_setup', 'tests']),
    include_package_data=True,
    scripts=['gprof2dot.py'],
    zip_safe=True,
    install_requires=install_requires,
)



# encoding: utf-8
# Created by Jeremy Bowman on Thu Feb 20 16:55:15 EST 2014
# Copyright (c) 2014, 2015 Safari Books Online. All rights reserved.
#
# This software may be modified and distributed under the terms
# of the 3-clause BSD license.  See the LICENSE file for details.
"""
Yet Another Django Profiler middleware

A django profiling middeware that tries to consolidate the best features of
the ones that came before it.  Originally described at
http://blog.safariflow.com/2013/11/21/profiling-django-via-middleware/
"""

__version__ = (1, 1, 0)  # Update docs/CHANGELOG.rst if you increment the version
